# if present, the C extension takes precedence over this file during import

from backports.zstd._cffi import (
    ZSTD_CLEVEL_DEFAULT,
    ZSTD_btlazy2,
    ZSTD_btopt,
    ZSTD_btultra,
    ZSTD_btultra2,
    ZSTD_c_chainLog,
    ZSTD_c_checksumFlag,
    ZSTD_c_compressionLevel,
    ZSTD_c_contentSizeFlag,
    ZSTD_c_dictIDFlag,
    ZSTD_c_enableLongDistanceMatching,
    ZSTD_c_hashLog,
    ZSTD_c_jobSize,
    ZSTD_c_ldmBucketSizeLog,
    ZSTD_c_ldmHashLog,
    ZSTD_c_ldmHashRateLog,
    ZSTD_c_ldmMinMatch,
    ZSTD_c_minMatch,
    ZSTD_c_nbWorkers,
    ZSTD_c_overlapLog,
    ZSTD_c_searchLog,
    ZSTD_c_strategy,
    ZSTD_c_targetLength,
    ZSTD_c_windowLog,
    ZSTD_d_windowLogMax,
    ZSTD_dfast,
    ZSTD_DStreamOutSize,
    ZSTD_fast,
    ZSTD_greedy,
    ZSTD_lazy,
    ZSTD_lazy2,
    ZstdCompressor,
    ZstdDecompressor,
    ZstdDict,
    ZstdError,
    finalize_dict,
    get_frame_info,
    get_frame_size,
    get_param_bounds,
    set_parameter_types,
    train_dict,
    zstd_version,
    zstd_version_number,
)
